% Parent script to compute simulated CMG distributions
%
% Establishes CMG model and computes statistics over simulated CMG studies.
%
% 1) Define all model parameters
% 2) Visualize nominal ureter inflow rates
% 3) Model sensitivity analysis
% 4) Simulations across large range if bladder infusion rates
% 5) Compare model to experiment results
%
%
% Dependencies:
%   - Calls SimCMGStats to execute multiple CMG simulations. Function takes
%       the struct p, containing simulation parameters, including p.N,
%       which is the number of simulations to compute statistics over. Uses
%       the Statistics and Machine Learning Toolbox.
%   - Calls ImportFormattedExpCMGData to load in and properly format data
%       from Jaskowak & Danziger 2023 for comparison.
%
% 
%
% ZC Danziger, November 2023
% MATLAB Version: 9.10.0.1602886 (R2021a)
%



%% 1) Define Nominal CMG Paramters

% --- define physical parameters ---

% Danziger, ZC, Grill, WM. Estimating postvoid residual volume without
%   measuring residual bladder volume during serial cystometrograms. Am J
%   Physiol Renal Physiol. 2016;311 (2):F459-68
% Jaskowak, DJ, Danziger, ZC. Reflex voiding in rat occurs at consistent
%   bladder volume regardless of pressure or infusion rate. Neurourol
%   Urodyn. 2023; 10.1002/nau.25243

p.avgVE = 0.85;       % mean voiding efficiency (target)      [-]
p.stdVE = 0.15;       % standard deviation of VE              [-] 
p.limVE = [0, 1];     % fractions are bounded by 0 and 1

p.avgBC = 0.25;       % mean functional bladder capacity      [mL]
p.stdBC = 0.05;       % standard deviation of BC              [mL]
p.limBC = [0.03, inf];% bladders must hold a minimum of 30uL

p.avgQu = 0.0033;     % mean ureter inflow rate               [mL/min]
p.stdQu = 0.31/60/2;  % std ureter inflow rate (gamma)        [mL/min]

p.setQi   = 0.025;    % selected bladder infusion rate        [mL/min]
p.noiseQi = 0.001;    % infusion rate hardware noise          [mL/min]
p.limQi = [0, inf];   % pump infusion limits

p.noiseVv = 0.005;    % measurement noise on voided volume    [mL]
p.noiseVr = 0.05;     % measurement noise on residual volume  [mL]



% --- define simulation parameters ---

% validity bounds on ureter inflow rate
p.validQu = [0, p.avgQu+p.stdQu*6];
% number of CMGs to simulate (per call to SimCMGStats)
p.N       = 1e5;   




%% 2) Visualization of Nominal Parameterization

% run a simulation using the nominal parameters
[~, AllCMGs] = SimCMGStats(p);



% ==> display distributions of estimated and actual ureter inflow rates
figure

subplot(2,1,1)
histogram(AllCMGs{:,'Qu'},...
    'Normalization','probability','BinWidth',1e-3)
hold on
histogram(AllCMGs{:,'VuPHE'}./AllCMGs{:,'tBC'},...
    'Normalization','probability','BinWidth',1e-3)
xlabel('Qu (mL/min)'); ylabel('frequency'); box off
hl = line([p.validQu(1) p.validQu(1) nan p.validQu(2) p.validQu(2)],...
        [0 0.2 nan 0 0.2],'linestyle','-','color',[1 1 1]*0.75);
uistack(hl,'bottom')
legend('Range of Validity','Qu','Qu-PostHoc')

% ==> display distributions of estimated and actual voiding efficiencies
subplot(2,1,2)
errVEAll = AllCMGs{:,'VE'} - AllCMGs{:,'msrVv'}./(AllCMGs{:,'msrVv'}+AllCMGs{:,'msrVr'});
histogram( errVEAll,'Normalization','probability','BinWidth',0.01)
hold on
ixOut = (AllCMGs{:,'VuPHE'} <= 0) | (AllCMGs{:,'VuPHE'}./AllCMGs{:,'tBC'} >= p.avgQu+p.stdQu*6);
errVESafe = AllCMGs{~ixOut,'VE'} - AllCMGs{~ixOut,'msrVv'}./(AllCMGs{~ixOut,'msrVv'}+AllCMGs{~ixOut,'msrVr'});
histogram( errVESafe,'Normalization','probability','BinWidth',0.01)
legend("All Voids","Voids Passing Qu Check")
xlabel('Voiding Efficiency Error')
ylabel('frequency'); box off

% ==> some quantitation of mean absolute differences
disp("     All  |  Passing | Failing")
disp( cellfun(@(u) mean(abs(u)),{errVEAll; errVEAll(~ixOut); errVEAll(ixOut)})' )



%% 3) Run Simulation Sensitivity Analysis

% parameters to include in sensitivity analysis sweep
pNames = {'avgVE','stdVE','avgBC','stdBC','avgQu','stdQu',...
    'setQi','noiseQi','noiseVv','noiseVr'}';

% what combinations of perterbation percentages away from nominal should we test
parPerts = [-50:5:-10 10:5:50];

% Simulation with nominal values
[RMaster, AllCMGs] = SimCMGStats(p);

% OFAT sensitivity analysis
for pn = 1:length(pNames)
    % a fresh set of parameters
    pPerturbed = p;         
    for k = 1:length(parPerts)
        % perturb parameter pn by the kth perturbation amount
        pPerturbed.(pNames{pn}) = p.(pNames{pn})*((100+parPerts(k))/100);
        % perform the simulation 
        R = SimCMGStats(pPerturbed);
        % append to the master results table
        RMaster = [RMaster; R];
    end
    disp("Simulations for parameter " + pNames{pn} + " complete.")
end


% ==> Display OFAT Sensitivity Analysis
figure
cmp = colorcube(length(pNames)+1);

for k=1:length(pNames)
    ix = (1:length(parPerts)) + (k-1)*length(parPerts) + 1;
    plot(RMaster{ix,"AvgErrVEPass"},RMaster{ix,"AvgErrVEFail"},'.',...
        'color',cmp(k,:),'MarkerSize',20)
    hold on
end
line([0.05 0.25],[0.05 0.25],'linestyle','--','color',[1 1 1]*0.5)
xlabel('VE Error on Passing CMGs')
ylabel('VE Error on Failing CMGs')
title("Perturbations from " + min(parPerts) + " to " + max(parPerts) + "% Nominal")
legend([pNames;{"Equivalence"}])
grid on; box off; axis equal




%% 4) Simulation Sweep Across Infusion Rates
List_setQi      = logspace(log10(5e-3), log10(5), 100);     % run these infusion rates
List_noiseQi    = [1e-4, 1e-3, 0.66e-2, 0.33e-2, 1e-2];     % run these infusion variabilities

% Simulation with nominal values
RQi = SimCMGStats(p);

for k = 1:length(List_setQi)                    % for each infusion rate
    for j = 1:length(List_noiseQi)              % for each infusion noise level

        % set up the parameters for this run by adjusting from nominal
        pPerturbed = p;                         % nominal
        pPerturbed.setQi = List_setQi(k);       % infusion rate update
        pPerturbed.noiseQi = List_noiseQi(j);   % pump noise update

        % perform the simulation 
        R = SimCMGStats(pPerturbed);

        % append to the master results table
        RQi = [RQi; R];
    end

    if any(k==round(linspace(0,length(List_setQi),5)))
        % simulation progress
        disp("Infusion sweep simulation " + round(100*k/length(List_setQi)) + "% complete.")
    end
end


% ==> Display Qi Sweep Results
RQi_QiVals   = [RQi{:,'SimulationPars'}.setQi];     % set flow rates for each simulation
RQi_QiNoise  = [RQi{:,'SimulationPars'}.noiseQi];   % pump noise for each simulation
RQi_FracFlag = RQi{:,'FracCMGFlags'};               % fraction of CMGs flagged for rejection

figure
scatter(RQi_QiVals,RQi_FracFlag,10,RQi_QiNoise,'filled');
xlabel('Pump Infusion Rate (ml/min)')
ylabel('Fraction of Trials Flagged for Rejection')
set(gca,"XScale","Log"); grid on; colormap lines
xlim([min(List_setQi) max(List_setQi)])

hold on
line([2.4 10.8]/60,[0.1 0.1],'color','r')




%% 5) Simulation of Experiment Data Across Infusion Rates

% import all the experiment data:
ExpCMGs = ImportFormattedExpCMGData;

% flowrates to simulate
List_setQi = logspace(-2, 0.12, 50);  % run these infusion rates

% Adjust nominal parameters to match this particular experiment protocol.
% Many parameters are independent of the given study, and therefore we do
% not update these based on incoming data. Parameters such as infusion rate
% fluctuation, and measurement noise on the voided or residual volumes.
pExp = p;
% Take means and stds of reliable parameters directly from data.
pExp.avgVE = mean( ExpCMGs.Vv./(ExpCMGs.Vv + ExpCMGs.Vr) );
pExp.stdVE = std(  ExpCMGs.Vv./(ExpCMGs.Vv + ExpCMGs.Vr) );
pExp.avgBC = mean( ExpCMGs.Vv + ExpCMGs.Vr  );
pExp.stdBC = std(  ExpCMGs.Vv + ExpCMGs.Vr  );
% Set Qu parameters to the post-hoc experiment estimates
pExp.avgQu = 0.0098;
pExp.stdQu = 0.0002;
% Match the CMG rejection bounds on Qu with experiment.
pExp.validQu = ExpCMGs.Qlims;



% Simulation with baseline values
ExpSimQi = SimCMGStats(pExp);

for k = 1:length(List_setQi)                    % for each infusion rate
    
        % set up the parameters for this run by adjusting from Exp baseline
        pPerturbed = pExp;                      % nominal
        pPerturbed.setQi = List_setQi(k);       % infusion rate update

        % perform the simulation 
        R = SimCMGStats(pPerturbed);

        % append to the master results table
        ExpSimQi = [ExpSimQi; R];

    if any(k==round(linspace(0,length(List_setQi),5)))
        % simulation progress
        disp("Infusion sweep simulation " + round(100*k/length(List_setQi)) + "% complete.")
    end
end



% Generate a second set of simulations to match the binned Qi obtained in
% the experiment to create an apples-to-apples comparison of the Qu-posthoc
% distributions.
QiExpBinned = ExpCMGs.QiBinned(~isnan(ExpCMGs.rejFracAvg));
SimQuPH = cell(size(QiExpBinned));
SimRejFrac = nan(size(QiExpBinned));
for k=1:length(QiExpBinned)
    % set up the parameters for this run by adjusting from Exp baseline
    pPerturbed = pExp;                      % nominal
    pPerturbed.setQi = QiExpBinned(k);      % infusion rate update
    % perform the simulation
    [R, AllCMGs] = SimCMGStats(pPerturbed);
    % store QuPH
    SimQuPH{k} = AllCMGs{:,'VuPHE'}./AllCMGs{:,'tBC'};
    % store rejected fraction of trials
    SimRejFrac(k) = R{1,'FracCMGFlags'};
end

% compute some measure of the "typical" missmatch
SimMismatch = median( SimRejFrac - ExpCMGs.rejFracAvg(~isnan(ExpCMGs.rejFracAvg)) );
disp("Average simulation overestimate: " + round(SimMismatch*100) + "%")

% uncomment below to run normality tests
% [~,pval(1)] = lillietest(SimRejFrac);
% [~,pval(2)] = lillietest(ExpCMGs.rejFracAvg(~isnan(ExpCMGs.rejFracAvg)));




% ==> Compare simulation and experiment
figure

% estimated ureter inflow rate, simulated and measured
subplot(1,2,1)
edges = -0.15:0.01:0.15;
histogram([SimQuPH{:}],edges,'Normalization','probability')
hold on
line([ExpCMGs.Qlims([1 1]) nan ExpCMGs.Qlims([2 2])],[0 0.5 nan 0 0.5],...
    'color',[1 1 1]*0.5)
histogram(ExpCMGs.QuEst,edges,'Normalization','probability')
set(gca,'box','off')
xlabel('Qu_{est} (ml/min)')

% overaly sim f(Q) and actual f(Q)
subplot(1,2,2)
plot([ExpSimQi{:,'SimulationPars'}.setQi],ExpSimQi{:,'FracCMGFlags'},...
    'marker','o','LineStyle','none','MarkerSize',4)

hold on
errorbar(ExpCMGs.QiBinned,ExpCMGs.rejFracAvg,ExpCMGs.rejFracStd,...
    'o','MarkerFaceColor','auto');
set(gca,'box','off','xscale','log')
xlabel('Infusion Rate (mL/min)')
ylabel('Fraction of Rejected Trials')








%% 6) Estimate the Qu distribution parameters from data

% The approch is to search across the two Qu generating distribution
% parameters for something that makes the *post hoc* Qu look like the
% experiment data post hoc Qu.


% import all the experiment data:
ExpCMGs = ImportFormattedExpCMGData;

% Adjust nominal parameters to match this particular experiment protocol.
% Many parameters are independent of the given study, and therefore we do
% not update these based on incoming data. Parameters such as infusion rate
% fluctuation, and measurement noise on the voided or residual volumes.
pExp = p;
% Take means and stds of reliable parameters directly from data.
pExp.avgVE = mean( ExpCMGs.Vv./(ExpCMGs.Vv + ExpCMGs.Vr) );
pExp.stdVE = std(  ExpCMGs.Vv./(ExpCMGs.Vv + ExpCMGs.Vr) );
pExp.avgBC = mean( ExpCMGs.Vv + ExpCMGs.Vr  );
pExp.stdBC = std(  ExpCMGs.Vv + ExpCMGs.Vr  );
pExp.avgQu = mean( ExpCMGs.QuEst );
% Match the CMG rejection bounds on Qu with experiment.
pExp.limsQu = ExpCMGs.Qlims;
% Take a central flowrate for the fit
pExp.setQi = ExpCMGs.QiBinned(6);
% Lower samples for the sake of computational efficiency
pExp.n = 1e4;


% Arrange the experimental data to fit
edges = -0.15:0.01:0.15;
[ExpDataQuPH,edges] = histcounts(ExpCMGs.QuEst,edges,'Normalization','probability');

% Visualize parameter combinations
% a) construct parameter combinations
[avgQu, stdQu] = meshgrid(linspace(p.avgQu*0.5,p.avgQu*5,30),linspace(p.stdQu*0.005,p.stdQu*10,30));
% b) define a loss function
lossFun = @(x1,x2) sum((SimAndReturnQuPH(pExp,[x1; x2],edges) - ExpDataQuPH).^2);
% c) evaluate loss at all combinations
L = arrayfun(lossFun,avgQu,stdQu);
% d) plot results
figure
imagesc([max(avgQu) min(avgQu)],[min(stdQu) max(stdQu)],L)
xlabel('avgQu'); ylabel('stdQu')
colorbar







