function [R, AllCMGs] = SimCMGStats(p)
% Given a struct of parameters, p, simulate CMG statistics
%
% Strategy: Select from specified distributions for each micturition cycle
% 1) bladder capacity, 2) voiding efficiency, 3) infusion rate, 4) ureter
% inflow rate. From these, compute cycle time, voided volume, and residual
% volume. Apply simulated measurement noise to the measured voided and
% residual volumes. Finally, compute a post hoc estimate of the infused
% ureter volume from our measurements of voided and residual volumes,
% presuming our infused volume is exactly as specified.
%
% [R, AllCMGs] = SimCMGStats(p)
%
% p contains all simulation parameters. (struct)
% R contains summary results for the parameters in p. (table)
% AllCMGs contain simulated values for all CMGs. (table)
%
%
% ZC Danziger July 2023
% Requires Statistics and Machine Learning Toolbox (made with ver 12.3)
%

%% List of Statistics to Compute
% table variable names:
TVarNames = {'VEActual','MeanAbsErrVu','MeanErrPctBC',...
    'AvgErrVEPass','AvgErrVEFail','FracCMGFlags','SimulationPars'};

%% Check that all parameter means are within their allowable ranges
if p.avgBC < p.limBC(1) || p.avgBC > p.limBC(2) || ...
   p.avgVE < p.limVE(1) || p.avgVE > p.limVE(2) || ...
   p.setQi < p.limQi(1) || p.setQi > p.limQi(2)
    [R, AllCMGs] = terminateSimulation(p,TVarNames);
    return
end

%% Function Definitions

% define new functions to support proper means for the truncated normals
nf    = @(m,s,c) (c - m)/s;                 % normalizing shifted function
phi   = @(u) (1/sqrt(2*pi))*exp(-u.^2/2);   % exponential
tncdf = @(u) (1 + erf(u/sqrt(2)))/2;        % cdf of truncated normal

% mean of the truncated normal given target mean, standard deviations, and distribution bounds
muTruncNorm = @(m,s,b) m + (phi(nf(m,s,b(1))) - phi(nf(m,s,b(2))))/(tncdf(nf(m,s,b(2)))-tncdf(nf(m,s,b(1))))*s;


% Bladder Capacity distribution
xBC = fzero(@(x) muTruncNorm(x,p.stdBC,p.limBC)-p.avgBC,p.avgBC);
pdfBC = truncate(makedist("Normal","mu",xBC,"sigma",p.stdBC),p.limBC(1),p.limBC(2));
% Voiding Efficiency distribution
xVE = fzero(@(x) muTruncNorm(x,p.stdVE,p.limVE)-p.avgVE,p.avgVE);
if ~isfinite(xVE), [R, AllCMGs] = terminateSimulation(p,TVarNames); return; end      % if there is no solution to the VE distribution, skip this simulation
pdfVE = truncate(makedist("Normal","mu",xVE,"sigma",p.stdVE),p.limVE(1),p.limVE(2));
% Infusion Rate distribution  
xQi = fzero(@(x) muTruncNorm(x,p.noiseQi,p.limQi)-p.setQi,p.setQi);
pdfQi = truncate(makedist("Normal","mu",xQi,"sigma",p.noiseQi),p.limQi(1),p.limQi(2));
% Ureter Inflow Rate distribution (Qu will rarely be zero, so use gamma distribution)
pdfQu = makedist("Gamma",'a',p.avgQu^2/p.stdQu^2,'b',p.stdQu^2/p.avgQu);


%% Produce CMGs

% --- stipulated values ---
% obtain BC for this CMG
BC = random(pdfBC,p.N,1);
% obtain VE for this CMG
VE = random(pdfVE,p.N,1);
% obtain actual infusion rate
Qi = random(pdfQi,p.N,1);
% obtain actual ureter inflow rate
Qu = random(pdfQu,p.N,1);
% ---

% --- forced values ---
% With Qu and Qi determined, we can compute how long the fill is:
% t*Qu+t*Qi=BC, t = BC/(Qu+Qi)
tBC = BC./(Qi+Qu);

% With the fill time known, we can compute the volume contributions.
Vu = tBC.*Qu;
Vi = tBC.*Qi;

% With inflow volumes known, we can compute voided and residual volumes.
Vv = VE.*(Vi+Vu);
Vr = BC - Vv;
% ---

% --- measurements ---
% We may now consider the quantities we measure:
% (note that we will never measure a negative volume)
msrVv = max(0, randn(p.N,1)*p.noiseVv + Vv );
msrVr = max(0, randn(p.N,1)*p.noiseVr + Vr );
% ---

% --- post hoc estimates ---
% Given measured and expected quantities, we can compute the experimenter's
% post hoc estimate of Vu, Vu', which was not measured or set. The estimate
% is based on the conservation of volume expression: t*Qu + t*Qi = Vv + Vr.
% Therefore, Vu' = Vv + Vr - t*Qi.
VuPHE = msrVv + msrVr - tBC*p.setQi;
% note how much error we have in the post hoc estimate
errVu = VuPHE-Vu;
% ---



%% Flag Suspicious CMGs
% Throw a flag that the CMG is suspicious if we estimate ureter volume
% contribution is negative or if the computed ureter flow rate is far above
% our expectations.
ixFlag = (VuPHE <= p.validQu(1)) | (VuPHE./tBC >= p.validQu(2));

% Average absolute error in VE estimates on non-flagged voids
errVEPass = mean(abs( VE(~ixFlag) - msrVv(~ixFlag)./(msrVv(~ixFlag)+msrVr(~ixFlag)) ));

% Average absolute error in VE estimate on flagged voids
errVEFlag = mean(abs( VE(ixFlag) - msrVv(ixFlag)./(msrVv(ixFlag)+msrVr(ixFlag)) ));



%% Outputs
if nargout==2
    % package full output if requested
    AllCMGs = table(BC,VE,Qi,Qu,tBC,Vu,Vi,Vv,Vr,msrVv,msrVr,VuPHE);
end



% summary of simulations
R = table(mean(VE),mean(abs(errVu)),100*mean(abs(errVu)./BC),errVEPass,errVEFlag,sum(ixFlag)/p.N,p,...
    'VariableNames',TVarNames);





function [r, allsims] = terminateSimulation(p,n)
    % simulation failed, return empty values without errors
    allsims = table();
    r = table(nan,nan,nan,nan,nan,nan,p,...
        'VariableNames',n);
end

end







